/* PDB Viewer by saintly@innocent.com
 *
 * This program is designed to read a PalmPilot .PDB data file and
 * display its data.  It displays most of the info from the database
 * header and dumps all the records in hex/ASCII format.
 *
 * This software is provided under the terms of the GNU GPL license.
 * Feel free to modify the source for your own purpose and redistribute 
 * under the terms of that license.  To view the latest copy of the
 * license, visit http://www.gnu.org
 */

#include <stdio.h>
#include <time.h>

// The structure of a PDB file header, for reference only
typedef struct db_header {
  char db_name[32];
  char flags[2];
  char version[2];
  char c_time[4]; // creation time
  char m_time[4]; // modification time
  char b_time[4]; // backup time
  char m_num[4];  // modification number
  char app_off[4]; // app info offset
  char srt_off[4]; // sort info offset
  char type[4]; // Database ID
  char creator[4]; // Application ID
  char unique_ID[4]; 
  char next_rec_list[4];
  char num_records[2];
} db_hdr;

void print_seg(char *buffer, int start, int end, char *label, int ret, int t);
short get_short(char *buffer, int start);
int get_int(char *buffer, int start);
void dump(FILE *read, int s_offset, int e_offset);

#define NTEXT 0 // null-terminated
#define TEXT  1 // text, not terminated
#define SHORT 2 // 2-byte number
#define INT   3 // 4-byte number
#define CNUM  4 // Characters as ASCII values (printed decimal)
#define DATE  5 // 4-byte date
#define XNUM  6 // Characters as ASCII values (printed hex)
#define FLAGS 7 // Display in binary

#define ERR_CANTOPEN 1
#define ERR_SHORT    2 

#define PERLINE 16 // Number of bytes per line for the hex dump
#define DISPLAY_APP_INFO 0  // Display the app info if true

int main(int argc, char *argv[]) {
  int file;
  for(file=1;file<argc;file++) {
    switch(dump_db(argv[file])) { 
      case ERR_CANTOPEN:
        printf("Error reading [%s], file doesn't exist.\n",argv[file]); 
        break;
      case ERR_SHORT:
        printf("Error reading [%s], file too short.\n",argv[file]); 
        break;
    } 
  }
  return 1;
}

int dump_db(char *filename) {
  FILE *pdbfile;
  char buffer[256];
  int bytes_read,i,*record_offsets,*ids;
  unsigned short max_records, app_info;

  pdbfile = fopen(filename, "r");
  if(!pdbfile) { return ERR_CANTOPEN; }
  printf("Reading database %s...\n",filename);
  bytes_read = fread(&buffer, 78, 1, pdbfile);
  print_seg(buffer,0,32, "DB Name           : ",1, NTEXT);
  print_seg(buffer,32,34,"Flags             : ",1, FLAGS);
  print_seg(buffer,34,36,"Version           : ",1, SHORT);
  print_seg(buffer,60,64,"Type              : ",1, TEXT);
  print_seg(buffer,64,68,"Creator           : ",1, TEXT);
  print_seg(buffer,36,40,"Creation time     : ",1, DATE);
  print_seg(buffer,40,44,"Modification time : ",1, DATE);
  print_seg(buffer,44,48,"Backup time       : ",1, DATE);
  if(DISPLAY_APP_INFO) {
    print_seg(buffer,52,56,"App info offset   : ",1, INT);
    print_seg(buffer,56,60,"Sort info offset  : ",1, INT);
  }
  print_seg(buffer,76,78,"Num. Records      : ",1, SHORT);

  app_info       = (unsigned int)get_int(buffer, 52);
  max_records    = (unsigned short)get_short(buffer,76);
  record_offsets = (int *)calloc(max_records+1, sizeof(int));
  ids            = (int *)calloc(max_records+1, sizeof(int));

  for(i=0;i<max_records;i++) { 
    bytes_read = fread(&buffer,8,1,pdbfile);
    if(bytes_read < 1) { fclose(pdbfile); return ERR_SHORT; }
    record_offsets[i] = (unsigned int)get_int(buffer,0);
               ids[i] = (unsigned int)get_int(buffer,4);
  }
  
  if(DISPLAY_APP_INFO) {
    printf("AppInfo [%d]\n",(unsigned int)app_info);
    dump(pdbfile, app_info, record_offsets[0]);
  }

  fseek(pdbfile,0,SEEK_END);
  record_offsets[max_records] = ftell(pdbfile)+1;
  for(i=0;i<max_records;i++) { 
    printf("Record# %d, ID %d (0x%X), flags (%0.2X), length %d, offset: %d\n",
        (unsigned int)i, (unsigned int)ids[i] & 0x00ffffff, 
        (unsigned int)ids[i] & 0x00ffffff, 
        (unsigned int)ids[i] >> 24, 
        (unsigned int)record_offsets[i+1] - (unsigned int)record_offsets[i],
        (unsigned int)record_offsets[i]);
    dump(pdbfile, record_offsets[i], record_offsets[i+1]);
  }

  fclose(pdbfile); return 0;
}

void dump(FILE *read, int s_offset, int e_offset) {
  unsigned char buf[PERLINE],i,j;
  if(fseek(read, s_offset,SEEK_SET) < 0) return;
  while((ftell(read) < e_offset) && !feof(read)) {
    printf( "%5d | ",(unsigned long)ftell(read) );
    i=0; while(i<PERLINE && !feof(read) && (ftell(read) < e_offset)) {
      if(fread(&buf[i],sizeof(char),1,read)) i++;
    }
    for(j=0;j<i;j++) { printf("%0.2X ",(unsigned char)buf[j]); }
    for(j=i;j<PERLINE;j++) { printf(".. "); }
    printf("|");
    for(j=0;j<i;j++) { printf("%c",(buf[j]<32 || buf[j]>126) ? '.' : buf[j]) ; }
    for(j=i;j<PERLINE;j++) { printf(" "); }
    printf("|\n");
  }
  // printf("\n");
}


short get_short(char *buffer, int start) {
   short n_buf;
   n_buf = ((unsigned char)buffer[start] << 8) + (unsigned char)buffer[start+1];
   return n_buf;
}

int get_int(char *buffer, int start) {
   short n_buf;
   n_buf = ((unsigned char)buffer[start] << 24) + 
           ((unsigned char)buffer[start+1] << 16) +
           ((unsigned char)buffer[start+2] << 8) +
           ((unsigned char)buffer[start+3]);
   return n_buf;
}

/* This function prints some data from a buffer, arguments are:
 *  *buffer (pointer to the buffer)
 *  start   (starting offset)
 *  end     (ending offset)
 *  label   (print this label before printing data)
 *  ret     (prints a newline after the data if true)
 *  t       (type of data from the #defines above)
 *
 */ 
void print_seg(char *buffer, int start, int end, char *label, int ret, int t) {
  int i,j; long n_buf; struct tm *dtime;
  if(label) { printf("%s[",label); }
  switch(t) { 
    case NTEXT: 
      for(i=start;i<end && buffer[i];i++) printf("%c",buffer[i]); 
      break;
    case TEXT:
      for(i=start;i<end;i++) printf("%c",buffer[i]); 
      break;
    case CNUM:
      for(i=start;i<end;i++) 
        printf("%s%d",(i!=start)?",":"",(unsigned char)buffer[i]); 
      break;
    case XNUM:
      for(i=start;i<end;i++) 
        printf("%s%0.2X",(i!=start)?",":"",(unsigned char)buffer[i]); 
      break;
    case FLAGS:
      for(i=start;i<end;i++) {
        printf("%s",(i!=start)?" ":"");
        for(j=0x80;j;j>>=1) { printf("%c",((unsigned char)buffer[i] & j) ? '1' : '0'); }
      }
      break;
    case SHORT:
      for(i=start;i<end;i+=2) {
        n_buf = ((unsigned char)buffer[i] << 8) + (unsigned char)buffer[i+1];
        printf("%u",n_buf);
      }
      break;
    case INT:
      for(i=start;i<end;i+=4) {
        n_buf = ((unsigned char)buffer[i] << 24) + 
                ((unsigned char)buffer[i+1] << 16) +
                ((unsigned char)buffer[i+2] << 8) +
                ((unsigned char)buffer[i+3]);
        printf("%u",n_buf);
      }
      break;
    case DATE:
      for(i=start;i<end;i+=4) {
        n_buf = ((unsigned char)buffer[i] << 24) + 
                ((unsigned char)buffer[i+1] << 16) +
                ((unsigned char)buffer[i+2] << 8) +
                ((unsigned char)buffer[i+3]);
        dtime = localtime(&n_buf);
        if(n_buf) { 
          printf("%d/%d/%d, %d:%d:%d",
               (dtime->tm_mon)+1,dtime->tm_mday,dtime->tm_year+1900,
               dtime->tm_hour,dtime->tm_min,dtime->tm_sec);
        } else { printf(" EMPTY "); }
      }
      break;
  }
  printf("]");
  if(ret) { printf("\n"); }
}
